

URDF_CONVERSION_LOG_LEVEL_NAME = "BROM_URDF_CONVERSION"
URDF_CONVERSION_LEVEL = 21

def does_drake_parser_support(filename: str):
    """
    Description
    -----------
    This function cleanly answers whether the given 3d object
    file is supported by Drake.
    :param filename:
    :return:
    """
    return ".obj" in filename # TODO(kwesi): Determine if .sdf should be put here.


from importlib import resources as impresources

from pydrake.geometry import SceneGraph
from pydrake.math import RollPitchYaw, RigidTransform, RotationMatrix
from pydrake.multibody.parsing import Parser
from pydrake.multibody.plant import MultibodyPlant
from pydrake.multibody.tree import FixedOffsetFrame
from pydrake.systems.framework import Diagram, DiagramBuilder
from pydrake.systems.primitives import Demultiplexer

# Local imports
from .gripper_type import GripperType
from ..control.cartesian_arm_controller import CartesianArmController

from brom_drake import robots


class UR10eStation(Diagram):
    """
    A template system diagram for controlling a UR10e robot in a simulated environment.
    """

    def __init__(
        self,
        time_step: float = 0.002,
    ):
        """
        Description:

            This class defines the
        """
        # Input Processing
        self.arm_urdf_path = str(
            impresources.files(robots) / "models/ur/ur10e.urdf",
        )

        # Initialize the Diagram
        Diagram.__init__(self)
        self.set_name("UR10e_Station")

        # Create a diagram builder
        self.builder = DiagramBuilder()

        # Create scene_graph and plant
        self.scene_graph = self.builder.AddSystem(SceneGraph())
        self.scene_graph.set_name("UR10e_Station_SceneGraph")

        self.plant = self.builder.AddSystem(
            MultibodyPlant(time_step=time_step)
        )
        self.plant.RegisterAsSourceForSceneGraph(self.scene_graph)
        self.plant.set_name("UR10e_Station_Plant")

        # Add a shadow plant that only has access to the robot arm + gripper mass,
        # and is used for the controller. (Doesn't know about anything else in the scene)
        self.controller_plant = MultibodyPlant(time_step=time_step)

        # Body ID's and Poses for Anything else in the scene
        self.object_ids = []
        self.object_poses = []

        # Which sort of gripper we're using (if any)
        self.gripper_type = GripperType.NoGripper

        # Whether we have a camera in the simulation
        # self.has_camera = False

        self.AddArm()

    def AddArm(self):
        """
        Add the UR10e arm to the system.
        """
        # Setup
        arm_urdf = self.arm_urdf_path

        # The hardware system has lots of damping so this is more realistic,
        # but requires a simulation with small timesteps.

        self.arm = Parser(plant=self.plant).AddModels(arm_urdf)
        self.controller_arm = Parser(plant=self.controller_plant).AddModels(arm_urdf)

        # Fix the base of the arm to the world
        self.plant.WeldFrames(
            self.plant.world_frame(),
            self.plant.GetFrameByName("base_link", self.arm),
        )

        self.controller_plant.WeldFrames(
            self.controller_plant.world_frame(),
            self.controller_plant.GetFrameByName("base_link", self.controller_arm),
        )

        # Create a new frame with the actual end-effector position.
        self.X_ee = RigidTransform()
        self.X_ee.set_translation([0, 0, 0.13])
        self.plant.AddFrame(FixedOffsetFrame(
            "end_effector",
            self.plant.GetFrameByName("end_effector_link"),
            self.X_ee, self.arm))
        self.controller_plant.AddFrame(FixedOffsetFrame(
            "end_effector",
            self.controller_plant.GetFrameByName("end_effector_link"),
            self.X_ee, self.controller_arm))

    def Finalize(self):
        """
        Description
        ----------
        This finalizes the diagram by:
        - Finalizing all plants in the diagram
        :return: Nothing
        """
        # Setup

        # Finalize all plants
        self.plant.Finalize()
        self.controller_plant.Finalize()

        # Set up the scene graph
        self.builder.Connect(
            self.scene_graph.get_query_output_port(),
            self.plant.get_geometry_query_input_port()
        )

        self.builder.Connect(
            self.plant.get_geometry_poses_output_port(),
            self.scene_graph.get_source_pose_port(self.plant.get_source_id())
        )

        # Create controller that uses the shadow plant to control the robot
        cartesian_controller = self.builder.AddSystem(
            CartesianArmController(self.controller_plant, self.controller_arm),
        )
        cartesian_controller.set_name("UR10e_Station_CartesianController")

        # End effector target and target type go to the controller
        self.builder.ExportInput(
            cartesian_controller.ee_target_port,
            "ee_target",
        )
        self.builder.ExportInput(
            cartesian_controller.target_type_port,
            "ee_target_type",
        )

        # Output measured arm position and velocity
        demux = self.builder.AddSystem(
            Demultiplexer(self.plant.num_multibody_states(self.arm), self.plant.num_positions(self.arm)),
        )
        demux.set_name("UR10e_Station_Demux")

        self.builder.Connect(
            self.plant.get_state_output_port(self.arm),
            demux.get_input_port(0),
        )
        self.builder.ExportOutput(
            demux.get_output_port(0),
            "measured_arm_position",
        )
        self.builder.ExportOutput(
            demux.get_output_port(1),
            "measured_arm_velocity",
        )

        # Measured Arm Position and Velocity are sent to the controller
        self.builder.Connect(
            demux.get_output_port(0),
            cartesian_controller.model_position_port,
        )
        self.builder.Connect(
            demux.get_output_port(1),
            cartesian_controller.model_velocity_port,
        )

        # Torques from controller go to the "true" plant
        self.builder.Connect(
            cartesian_controller.control_output_port,
            self.plant.get_actuation_input_port(self.arm),
        )

        # Controller outputs measured arm torques, end-effector pose, end-effector twist
        controller_output_port_names = ["applied_arm_torque", "measured_ee_pose", "measured_ee_twist"]
        for output_port_name in controller_output_port_names:
            self.builder.ExportOutput(
                cartesian_controller.GetOutputPort(output_port_name),
                output_port_name,
            )

        # Create gripper controller


"""
Description
-----------
This file contains the URDF converter for transforming
arbitrary URDFs into URDFs that the Drake toolbox can
support.
"""
import os
from copy import copy, deepcopy
from typing import Tuple

import trimesh
from doctest import UnexpectedException

import loguru
from multiprocessing.managers import Value
from pathlib import Path
import xml.etree.ElementTree as ET

# Internal Supports
from .util import (
    does_drake_parser_support,
    URDF_CONVERSION_LOG_LEVEL_NAME, URDF_CONVERSION_LEVEL
)


class DrakeReadyURDFConverter:
    def __init__(
        self,
        original_urdf_filename: str,
        output_urdf_file_path: str = None,
        overwrite_old_models: bool = False,
        overwrite_old_logs: bool = False,
        log_file_name: str = "conversion.log",
    ):
        # Setup
        self.original_urdf_filename = original_urdf_filename
        self.output_urdf_file_path = output_urdf_file_path
        self.log_file_name = log_file_name

        # Define conversion data directory in brom
        # (If it already exists, then we will overwrite it with the proper flag set)
        self.models_dir = Path("./brom/models/")
        if overwrite_old_logs and os.path.exists(self.models_dir / self.log_file_name):
            os.remove(self.models_dir / self.log_file_name)
        if overwrite_old_models:
            self.clean_up_models_dir()
        os.makedirs(self.models_dir, exist_ok=True)

        # Input Processing
        original_urdf_path = Path(self.original_urdf_filename)
        assert original_urdf_path.exists(), f"URDF file {original_urdf_path} does not exist!"

        if not (".urdf" in original_urdf_path.name):
            raise ValueError("Original URDF file must have a .urdf extension!")

        # Set Up logger
        self.configure_logger()
        DrakeReadyURDFConverter.log(
            f"Created a new DrakeReadyURDFConverter for file \"{self.original_urdf_filename}\"."
        )

    def configure_logger(self):
        """
        Description
        -----------
        This method configures the loguru logger for the URDF conversion.
        :return:
        """
        # Setup
        urdf_conversion_level_exists = False
        log_file_name = self.log_file_name

        # Check to see if the log level exists
        try:
            found_level = loguru.logger.level(URDF_CONVERSION_LOG_LEVEL_NAME)
            urdf_conversion_level_exists = (found_level.name == URDF_CONVERSION_LOG_LEVEL_NAME)
        except Exception as e:
            # If the logger doesn't exist, then we will create it.
            expected_message = f"Level '{URDF_CONVERSION_LOG_LEVEL_NAME}' does not exist"
            if expected_message in str(e):
                pass
            else:
                raise UnexpectedException(f"Unexpected exception: {e}")

        # Configure logger if it doesn't exist
        if urdf_conversion_level_exists:
            loguru.logger.add(
                self.models_dir / log_file_name,
                level=URDF_CONVERSION_LOG_LEVEL_NAME,
                filter=lambda record: record['level'].name == URDF_CONVERSION_LOG_LEVEL_NAME,
            )
        else:
            loguru.logger.level(
                URDF_CONVERSION_LOG_LEVEL_NAME,
                no=URDF_CONVERSION_LEVEL,
            )
            loguru.logger.add(
                self.models_dir / log_file_name,
                level=URDF_CONVERSION_LOG_LEVEL_NAME,
                filter=lambda record: record['level'].name == URDF_CONVERSION_LOG_LEVEL_NAME,
            )

    def convert_urdf(self) -> Path:
        """
        Description
        -----------
        Converts the original URDF file into a Drake-compatible URDF file.
        """
        # Setup
        original_urdf_path = Path(self.original_urdf_filename)
        original_xml = ET.ElementTree(file=original_urdf_path)

        # Use recursive function to convert the tree
        new_tree = self.convert_tree(original_xml)

        # Output the new tree to a file
        output_urdf_path = self.output_file_directory() / self.output_urdf_file_name()
        os.makedirs(output_urdf_path.parent, exist_ok=True)

        new_tree.write(output_urdf_path)

        DrakeReadyURDFConverter.log(
            f"Converted URDF file to Drake-compatible URDF file at {output_urdf_path}.\n\n"
        )
        return output_urdf_path

    def clean_up_models_dir(self):
        """
        Description
        -----------
        This method will clean up the models directory.
        :return:
        """
        # Make sure that the models directory exists
        os.makedirs(self.models_dir, exist_ok=True)
        os.system(f"rm -r {self.models_dir}")

    def convert_tree(
        self,
        current_tree: ET.ElementTree
    ) -> ET.ElementTree:
        """
        Description
        -----------
        This recursive method is used to parse each element of the
        xml element tree.
        :param current_tree:
        :return:
        """
        # Setup
        root = deepcopy(current_tree.getroot())

        # Modify Root if it contains a mesh or file input
        new_root = self.convert_tree_element(root)

        # Remove all children from the new root
        for child in new_root.findall("*"):
            new_root.remove(child)

        # Modify Children
        for child in root:
            new_child = self.convert_tree(
                ET.ElementTree(child)
            )
            new_root.append(new_child.getroot())

        return ET.ElementTree(new_root)


    def convert_tree_element(
        self,
        elt: ET.Element,
    ) -> ET.Element:
        """
        Description
        -----------
        This method is used to transform one element in the XML tree into an eleemnt
        that that Drake URDF parser can handle. If necessary, it will create a new
        3d model file and place it into the right directory.
        :param elt: An element in the ElementTree that we would like to convert
                     into a Drake-ready element.
        :return: The Drake-ready xml tree element.
        """
        # Setup
        new_elt = deepcopy(elt)

        # Convert file paths for mesh elements if necessary
        if new_elt.tag == "mesh":
            self.log(f"Found a mesh element with filename \"{elt.attrib['filename']}\".")
            # Check the value of the filename
            if not ("filename" in new_elt.attrib):
                raise ValueError(
                    f"Found a mesh element that does not contain the \"filename\" attribute.\n"
                    +"Brom doesn't know how to handle this!"
                )

            # Filename exists; Let's check to see if it's obj or not
            if does_drake_parser_support(new_elt.attrib["filename"]):
                pass
            else:
                # If parser does not support the given filename,
                # then let's try to create one
                old_filename = elt.attrib["filename"]
                new_filename = self.create_obj_to_replace_mesh_file(old_filename)
                new_elt.set(
                    "filename", new_filename,
                )
                self.log(
                    f"Replaced the mesh file \"{old_filename}\" with a Drake-compatible .obj file at \"{new_filename}\"."
                )
        else:
            # If the element is not a mesh element, then we will just return a copy of it.
            self.log(
                f"Found an element with tag \"{elt.tag}\" that is not a mesh element. No modification needed..."
            )
        # Otherwise, just return a copy of the previous element.
        return new_elt

    def create_obj_to_replace_mesh_file(
        self,
        mesh_file_name: str,
    ) -> str:
        """
        Description
        -----------
        This function will create an .obj file to replace the
        .stl or .dae file that is given in mesh file "mesh_file_name".
        :param mesh_file_name:
        :return:
        """
        # Setup
        original_urdf_dir = Path(self.original_urdf_filename).parent

        # Find the location of the mesh_file_name relative to the urdf file
        original_file_location_relative_to_urdf = None
        if mesh_file_name.startswith("./"):
            self.log(f"Found a mesh file with a relative path: {mesh_file_name}")
            original_file_location_relative_to_urdf = Path(mesh_file_name).parent
        elif mesh_file_name.startswith("package:"):
            self.log(f"Found a mesh file with a package path: {mesh_file_name}")
            _, original_file_location_relative_to_urdf = self.find_file_path_in_package(mesh_file_name)

            self.log(f"File path should be {original_file_location_relative_to_urdf}")
        else:
            raise ValueError(
                f"Mesh file path {mesh_file_name} is not supported by this function!\n" +
                "Make sure that the \"filename\" attribute contains a relative path (i.e., starts with \"./\")\n"
                " or a package path (i.e., starts with \"package:\")."
            )

        # Define the paths to each file
        directory_for_transformed_file = original_file_location_relative_to_urdf

        # Make sure that the file's location exists

        # Convert based on the file type

        # First check to see if the file type is supported
        output_path_for_urdf = None

        print(f"Original urdf dir: {original_urdf_dir}")
        print(f"Original file location relative to urdf: {original_file_location_relative_to_urdf}")
        print(f"Mesh file name: {Path(mesh_file_name).name}")

        print(f"original_urdf_dir / original_file_location_relative_to_urdf = {original_urdf_dir / original_file_location_relative_to_urdf}")
        original_file_location = original_urdf_dir / original_file_location_relative_to_urdf / Path(mesh_file_name).name

        print(f"Original file location: {original_file_location}")

        # Check to see if the file type is supported
        supported_suffixes = [".stl", ".dae", ".DAE", ".STL"]
        contains_supported_suffix = False
        for suffix in supported_suffixes:
            if suffix in mesh_file_name:
                output_path_for_urdf = directory_for_transformed_file / Path(mesh_file_name).name.replace(suffix, ".obj")
                break

        # If the file type is supported, then output_path_for_urdf should be defined
        if output_path_for_urdf is None:
            raise ValueError(f"File type {mesh_file_name} is not supported by this function!")

        # Convert the file
        new_mesh = trimesh.load(str(original_file_location))

        # Write the new mesh to the output path
        output_path_for_exporter = self.output_file_directory() / output_path_for_urdf
        os.makedirs(output_path_for_exporter.parent, exist_ok=True)
        new_mesh.export(str(output_path_for_exporter))

        return "./" + str(output_path_for_urdf)

    def find_file_path_in_package(
            self,
            file_path: str,
            max_depth: int = 10,
    ) -> Tuple[Path, Path]:
        """
        Description
        -----------
        This method will find the file path in the package.
        :param file_path: The raw falue of the "filename" attribute in the mesh element.
        :param max_depth: The maximum depth to search for the package path.
        :return:
        """
        # Setup
        original_urdf_path = Path(self.original_urdf_filename)

        # Find the path to the package
        package_found = False
        candidate_path = original_urdf_path.parent
        search_depth = 1
        while not package_found:
            # Check to see if "package.xml" exists in the directory
            if (candidate_path / "package.xml").exists():
                self.log(f"Found package path at {candidate_path}.")
                package_found = True
            else:
                candidate_path = candidate_path.parent
                search_depth += 1

            if search_depth > max_depth:
                raise ValueError(
                    f"Could not find package path for file {file_path} after searching {max_depth} directories."
                )

        # Should now have found the package directory
        package_dir = candidate_path

        # Open the package.xml file and find the name of the package
        package_xml = ET.ElementTree(file=package_dir / "package.xml")
        package_root = package_xml.getroot()
        package_name = package_root.find("name").text

        self.log(f"Found package name: {package_name}")

        # Create 2 Outputs:
        # 1. path to file by replacing the package name with the package path
        full_path_to_file = package_dir / Path(file_path.replace(f"package://{package_name}", "")).parent
        # 2. Relative path to the file from the urdf directory
        path_from_urdf_to_file = Path(".")
        for ii in range(search_depth-1):
            path_from_urdf_to_file = path_from_urdf_to_file / Path("..")
        path_from_urdf_to_file = path_from_urdf_to_file / Path(file_path.replace(f"package://{package_name}/", "")).parent

        return full_path_to_file, path_from_urdf_to_file

    @staticmethod
    def log(message: str):
        """
        Description
        -----------
        Logs a message to the logger.
        :param message: A string with the message we want to send to the logs.
        :return:
        """
        loguru.logger.log(URDF_CONVERSION_LOG_LEVEL_NAME, message)

    def output_file_directory(self) -> Path:
        """
        Description
        -----------
        Returns the directory of the output urdf file.
        :return:
        """
        # Setup
        original_file_path = Path(self.original_urdf_filename)

        # Input Processing
        if self.output_urdf_file_path is None:
            return self.models_dir / original_file_path.name.replace(".urdf", "")
        else:
            output_file_path = Path(self.output_urdf_file_path)
            return output_file_path.parent

    def output_urdf_file_name(self) -> str:
        """
        Description
        -----------
        Creates the output filename based on the original filename.
        """
        # Setup
        original_file_path = Path(self.original_urdf_filename)

        if self.output_urdf_file_path is None:
            # Use Pathlib to extract the filename, if it exists
            original_name = original_file_path.name
            original_name = original_name.replace(".urdf", "")

            return f"{original_name}.drake.urdf"
        else:
            return Path(self.output_urdf_file_path).name


"""
PortWatcher.py
Description:

    This file defines the PortWatcher class. This class is used to watch the ports of a
    diagram.
"""
from typing import List, Tuple, Union, NamedTuple
import loguru
import numpy as np
import matplotlib.pyplot as plt
import os

from pydrake.systems.framework import OutputPort, PortDataType, DiagramBuilder, LeafSystem
from pydrake.systems.primitives import LogVectorOutput
from pydrake.systems.framework import Context

# Internal Imports
from .PortWatcherOptions import PortWatcherOptions, PortFigureArrangement

class PortWatcher:
    def __init__(
        self,
        system: LeafSystem,
        output_port: OutputPort,
        builder: DiagramBuilder,
        logger_name: str = None,
        options: PortWatcherOptions = PortWatcherOptions(),
        plot_dir: str = "./brom",
    ):
        # Setup
        self.options = options
        self.system = system
        self.port = output_port
        self.data = {}
        self.plot_handles = {}
        self.plot_handles = None
        self.plot_dir = plot_dir

        # Input Processing
        if output_port.get_data_type() != PortDataType.kVectorValued:
            raise ValueError(
                f"This watcher only supports vector valued ports (i.e., of type {PortDataType.kVectorValued}.\n" +
                f"Received port of type {output_port.get_data_type()}."
            )

        if logger_name is None:
            logger_name = f"PortWatcher_{system.get_name()}_{output_port.get_name()}"

        # Processing
        self.logger = LogVectorOutput(output_port, builder)
        self.logger.set_name(logger_name)

    def plot_logger_data(
        self,
        diagram_context: Context,
    ) -> Tuple[List[plt.Figure], List[List[plt.Axes]]]:
        """
        plot_logger_data
        Description:

            This function plots the data in the logger.
        :param diagram_context:
        :return:
        """
        # Setup

        # Get the log from the logger
        temp_log = self.logger.FindLog(diagram_context)

        log_times = temp_log.sample_times()
        log_data = temp_log.data()

        if (log_data.shape[1] == 0) or (log_data.shape[0] == 0):
            loguru.logger.warning(
                f"No data found for {self.system.get_name()} - Port {self.port.get_name()}! Skipping...")
            return None, None

        # Plot the data
        if self.options.plot_arrangement == PortFigureArrangement.OnePlotPerPort:
            fig, ax_list = self.plot_logger_data_subplots(diagram_context, log_times, log_data)
            return [fig], [ax_list]

        elif self.options.plot_arrangement == PortFigureArrangement.OnePlotPerDim:
            figs, ax_grid = [], []
            for port_index in range(self.port.size()):
                fig_ii = plt.figure()
                ax_ii = fig_ii.add_subplot(1, 1, 1)
                ax_ii.plot(
                    log_times, log_data[port_index, :]
                )

                # Save figures and axes to lists
                figs.append(fig_ii)
                ax_grid.append([ax_ii])

            return figs, ax_grid

        else:
            raise ValueError(
                f"Invalid plot arrangement: {self.options.plot_arrangement}."
            )

    def plot_logger_data_subplots(
        self,
        diagram_context: Context,
        times: np.array,
        data: np.array,

    ):
        """
        plot_logger_data_subplots
        Description:

            This function plots the data in the logger.
        :param times:
        :param data:
        :param diagram_context:
        :return:
        """
        # Setup
        n_dims = data.shape[0]

        # Plot the data
        n_rows, n_cols = self.compute_plot_shape(n_dims)


        fig, ax_list = plt.subplots(n_rows, n_cols)

        for row_index in range(n_rows):
            for col_index in range(n_cols):

                dim_index = n_cols * row_index + col_index

                if dim_index >= n_dims:
                    fig.delaxes(ax_list[row_index, col_index])
                    continue

                ax_list[row_index, col_index].plot(times, data[dim_index, :])
                ax_list[row_index, col_index].set_title(f"Dim #{dim_index}")

        return fig, ax_list

    def compute_plot_shape(self, n_dims: int) -> Tuple[int, int]:
        """
        Description:
            Computes the shape of the plot based on the data.
        :param n_dims: The number of dimensions in the data.
        :return:
        """
        if n_dims == 1:
            return 1, 1
        if n_dims == 2:
            return 1, 2

        if n_dims < 9:
            return 2, int(np.ceil(n_dims / 2.0))

        # Otherwise
        return 3, int(np.ceil(n_dims / 3.0))

    def safe_system_name(self) -> str:
        """
        Description:
            Returns a safe name for the system.
        :param name: System's name.
        :return:
        """
        out = self.system.get_name()
        # First, let's check to see how many "/" exist in the name
        slash_occurences = [i for i, letter in enumerate(out) if letter == "/"]
        if len(slash_occurences) > 0:
            out = out[slash_occurences[-1] + 1:]  # truncrate string based on the last slash

        # Second, replace all spaces with underscores
        out = out.replace(" ", "_")

        return out

    def save_figures(self, diagram_context: Context):
        """
        savefigs
        Description:

            This function saves the figures.
        :param diagram_context:
        :return:
        """
        # Setup
        figs, ax_list = self.plot_logger_data(diagram_context)

        os.makedirs(self.plot_dir, exist_ok=True)

        # Save the figures (if possible)
        if figs is None:
            return

        if len(figs) == 1:
            figs[0].savefig(
                f"{self.plot_dir}/{self.safe_system_name()}_{self.port.get_name()}.png",
                dpi=self.options.plot_dpi,
            )
        else:
            # Create a directory for the plots
            port_plot_dir = self.plot_dir + f"/{self.safe_system_name()}_{self.port.get_name()}"
            os.makedirs(port_plot_dir, exist_ok=True)

            # Plot each figure within this directory
            for ii, fig_ii in enumerate(figs):
                fig_ii.savefig(
                    f"{port_plot_dir}/dim{ii}.png",
                    dpi=self.options.plot_dpi,
                )



"""
PortWatcherOptions.py
Description:

    This class defines the options used when making a PortWatcher object.
"""

from enum import Enum
from typing import NamedTuple


class PortFigureArrangement(Enum):
    OnePlotPerDim = 0
    OnePlotPerPort = 1


class PortWatcherOptions(NamedTuple):
    plot_arrangement: PortFigureArrangement = PortFigureArrangement.OnePlotPerPort
    plot_dpi: int = 300
    save_to_file: bool = True
